#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Feb 20 11:47:45 2020

@author: cdu022

The "Read" Oblo instrumentation files - function collection 
"""
from toblo import *
import pandas as pd
import numpy as np
import glob, os
import copy
import numpy.ma as ma 
from scipy.io import loadmat
import datetime as dt
from datetime import datetime, timedelta 
import netCDF4 as nc
from netCDF4 import Dataset
from netCDF4 import num2date, date2num 

# =============================================================================
###############################################################################
############################## IN SITU MESUREMENTS ############################
###############################################################################
# =============================================================================
# =============================================================================
# Read Alliance Ship meteorology data
# =============================================================================
def read_alliance(data_loc):
    """
    data_loc = '/scratch/Data/IGP/Alliance_Meteo_QC2/IGP_ship_met_QC2.nc'

    """
    dataset = Dataset(data_loc, 'r')

    time = pd.to_datetime(num2date(dataset['time'][:].data,'minutes since 2018-02-05 00:00:00'))
    sst  =     dataset['sea_surface_temperature_interpolated'][:].data
    airT =     dataset['air_temperature_interpolated'][:].data
    rh  =      dataset['relative_humidity'][:].data
    p  =       dataset['air_pressure'][:].data
    wind =     dataset['wind_speed'][:].data
    wind_dir = dataset['wind_direction'][:].data
    longitude = dataset['longitude_nmea'][:].data
    latitude = dataset['latitude_nmea'][:].data
    heading = dataset['heading_nmea'][:].data
    speed   = dataset['water_speed_nmea'][:].data
    ship_data = pd.DataFrame({'sst':sst,'airT':airT, 'rh':rh, 'p':p, 'wind':wind, 
                              'wind_dir':wind_dir, 'lon':longitude,
                              'lat':latitude, 'speed':speed,'heading':heading}, index = time)
    dataset.close()
    return ship_data

# =============================================================================
# Read the ship log from Kristine Bonnevie
# =============================================================================

def read_kristine_bonnevie(file):
    """
    file = '/scratch/Data/GEOF232_2020/Ship_log/raw/pos02-03-2020.csv'
    """
    
    log = pd.read_csv(file, skiprows=2, encoding = "ISO-8859-1", engine='python')
    
    # Get the date
    day = file.split('/')[-1][3:5]
    month = file.split('/')[-1][6:8]
    year = file.split('/')[-1][9:13]
    
    log.Time = pd.to_datetime(year+'-'+month+'-'+day+' '+log.Time)
    log.Time.iloc[-1] = log.Time.iloc[-1]+timedelta(days=1)
    
    log.index = log.Time
    
    # Get the longitude and latitude information
    lon = np.zeros(len(log))
    lat = np.zeros(len(log))
    for i in range(len(log)):
        try:
            lon[i] = float(log['Longitude'].iloc[i][:3]) + float((log['Longitude'].str.split(' ').iloc[i][0])[3:])/60.
            lat[i] = float(log['Latitude'].iloc[i][:2]) +  float((log['Latitude'].str.split(' ').iloc[i][0])[2:])/60.
        except TypeError:
            lon[i] = np.nan
            lat[i] = np.nan
            
    log['Longitude'] = lon
    log['Latitude']  = lat
    
    log[log==-999] = np.nan
    
    return log
    
# =============================================================================
# Read the Radiosonde Data (IGP, netcdf)
# =============================================================================

def read_radiosonde(data_loc):
    """
    
    """
    origin_dir = os.getcwd()        
    # Change to directory that includes the files to be read                                      
    os.chdir(data_loc)   

    # Scan list fo all dates of selected period (Must be slightly altered from original syntax)
    filenames = sorted(glob.glob('*.nc'))  
    #nc_data = Dataset(data_loc+'ncas-radiosonde-2_alliance_20180212-113916_sonde_v0.nc','r')

    sonde = {}
    for file in filenames:
        nc_data = Dataset(file)
        single  = {}
        single['temperature']       = nc_data['air_temperature'][:].data
        single['relative_humidity'] = nc_data['relative_humidity'][:].data
        single['pressure']          = nc_data['air_pressure'][:].data
        single['altitude']          = nc_data['altitude'][:].data
        single['time']              = pd.to_datetime(nc.num2date(nc_data['time'][:].data,
                                                               'seconds since 1970-01-01'))
        single['wind_speed']        = nc_data['wind_speed'][:].data
        single['wind_direction']    = nc_data['wind_from_direction'][:].data 
        sonde[file.split('_')[2]]   = pd.DataFrame(single, index = single['altitude'])
        nc_data.close()

    os.chdir(origin_dir) 

    temp = {}
    ws   = {}
    wd   = {}
    rh   = {}
    p    = {}
    dz  = 10
    
    ascents = []
    new_index = np.arange(10,4000,dz)
    for ascent in sonde.keys():
#        if ascent == '20180303-114432' or ascent == '20180305-175442' or ascent == '20180316-204817' or ascent =='20180216-114326':
#            print('skip')
#        else:
##            print(ascent)
        reindex = sonde[ascent].copy()
        reindex.index = sonde[ascent].index.astype(int)
        reindex = reindex[~reindex.index.duplicated()]
        reindex = reindex.sort_index()
        reindex = reindex.interpolate().reindex(index=new_index,method='bfill').interpolate()#np.array(windcube.level)+10
      
        # Combine the reindexed time series of 
        time_stamp = reindex.time[reindex.index[0]].strftime('%Y-%m-%d %H:%M:%S')
      
        temp[time_stamp] = np.array(reindex.temperature)
        ws[time_stamp]   = np.array(reindex.wind_speed)
        wd[time_stamp]   = np.array(reindex.wind_direction)
        rh[time_stamp]   = np.array(reindex.relative_humidity)
        p[time_stamp]    = np.array(reindex.pressure)
        del reindex
        ascents.append(ascent)
        
    temp = pd.DataFrame(temp, index = new_index)
    wd   = pd.DataFrame(wd, index = new_index)                    
    ws   = pd.DataFrame(ws, index = new_index)
    rh   = pd.DataFrame(rh, index = new_index)
    p    = pd.DataFrame(p, index = new_index)

    radio_ws = ws.transpose()
    radio_ws.index = pd.to_datetime(radio_ws.index)
    radio_wd = wd.transpose()
    radio_wd.index = pd.to_datetime(radio_wd.index)

    radio_temp = temp.transpose()
    radio_temp.index = pd.to_datetime(radio_temp.index)
    radio_rh = rh.transpose()
    radio_rh.index = pd.to_datetime(radio_rh.index)
    radio_p  = p.transpose()
    radio_p.index = pd.to_datetime(radio_p.index)    
    
    return radio_temp, radio_rh, radio_p, radio_ws, radio_wd

# =============================================================================
# Netcdf Datasets
# =============================================================================

def read_netcdf(file):
    """
    Read netcdf datasets (with maximum of 2 dimension variables)
    """
    # Define the file
    data = Dataset(file)

    # Get the time vector
    try:
        time = num2date(data['time'][:].data, units=data['time'].units,
                            calendar='gregorian')
    # In case the netcdf stars in year 0000 (not possible in netcdf4 module)
    except ValueError:
        # to unidata about this
        tunits = data['time'].units
        since_yr_idx = tunits.index('since ') + 6
        year = int(tunits[since_yr_idx:since_yr_idx+4])
        year_diff = year - 1

        new_units = tunits[:since_yr_idx] + '0001-01-01 00:00:00'
        time = num2date(data['time'][:].data, new_units, calendar='gregorian')
        time = [datetime(d.year + year_diff, d.month, d.day, 
                  d.hour, d.minute, d.second) for d in time]    

    # Get the variable names and dimensions
    var_name = []
    var_dim = []
        
    buffer = {}    
    for var in data.variables:
        var_name = np.append(var_name, var)
        var_dim = np.append(var_dim, len(data[var].shape))
        # Save the variable
        if var == 'level':
            buffer[var] = data[var][:].data
        elif len(data[var].shape) == 1:
            buffer[var] = pd.DataFrame(data[var][:].data, index = time, columns = [var])
        elif len(data[var].shape) == 2:
            buffer[var] = pd.DataFrame(data[var][:,:].data, index = time, columns = data['level'][:].data)
        
    data.close()
    return buffer

# =============================================================================
###############################################################################
################################ ALL LIDAR ####################################
###############################################################################
# =============================================================================


# =============================================================================
# Zephir 300 continuous wave Lidar
# =============================================================================

def read_zephir(path, delimiter, zephir_levels = [10, 20, 30, 38, 50, 60, 80, 100, 125, 150, 200], rot = 0, month = '_'):
    """
    Read the 3 dimentional wind speed components from the Zephir Lidar dataset (internally processed). 
    path = '/scratch/Data/Zephir/Wind_328@Y2020_M02_D20.ZPH.csv'
    delimiter = ','

    Parameters
    ==========
    
    path: str
        path to directory, where zephire files are stored
        
    delimiter: str (';',',')
        delimiter used in file (should be ',', however under certain conditions can also be ';'). Check in file.

    zephir_levels: list, default= [10, 20, 30, 38, 50, 60, 80, 100, 125, 150, 200]
        Levels choosen for the zephir Lidar to focus
    
    Returns
    =======
    
    uh: pandas dataframe
        dataframe containing horizontal wind speed at different heights
        
    wd: pandas dataframe
        dataframe containing wind direction at different heights
        
    w: pandas dataframe
        dataframe containing vertical wind speed at different heights 
        
    met: pandas dataframe
        dataframe containing all variables measured by the small attached weather station 
    
    """
    origin_dir = os.getcwd()        
    # Change to directory that includes the files to be read                                      
    os.chdir(path)   

    # Scan list fo all dates of selected period (Must be slightly altered from original syntax)
    filenames = sorted(glob.glob('*M'+month+'*.csv'))      
    

    # Get the column names
    uh_cols  = ['Horizontal Wind Speed (m/s) at '+str(h)+'m' for h in zephir_levels]    
    wd_cols  = ['Wind Direction (deg) at '+str(h)+'m' for h in zephir_levels]  
    w_cols   = ['Vertical Wind Speed (m/s) at '+str(h)+'m' for h in zephir_levels]
    met_cols = ['Met Tilt (deg)', 'Met Compass Bearing (deg)',  
                'Met Air Temp. (C)', 'Met Pressure (mbar)', 'Met Humidity (%)', 
                'Met Wind Speed (m/s)', 'Met Wind Direction (deg)']    
    pos_col = ['GPS']
    
    zephir = pd.DataFrame()
    for file in filenames:
        zephir_buf = pd.read_csv(file,delimiter=delimiter,skiprows=1)
        del zephir_buf[zephir_buf.columns[-1]]

        if delimiter == ';':
            # Replace all "," with "."
            for col in zephir_buf.columns[zephir_buf.columns.str.contains('Speed|Direction')]:
                zephir_buf[col] = zephir_buf[col].str.replace(",",".").astype(float, errors = 'ignore')

            for col in met_cols[1:-2]:    
                zephir_buf[col] = zephir_buf[col].str.replace(",",".").astype(float, errors = 'ignore')

            time_stamp = pd.to_datetime(zephir_buf['Time and Date'], format = '%d.%m.%Y %H:%M:%S')
        else:
            time_stamp = pd.to_datetime(zephir_buf['Time and Date'])
        zephir_buf.index = time_stamp
        zephir = pd.concat([zephir, zephir_buf], axis=0)

    uh = zephir[uh_cols].copy() 
    uh[uh>1000] = np.nan
    wd = zephir[wd_cols].copy() 
    wd[wd>1000] = np.nan
    w = zephir[w_cols].copy() 
    w[w>1000] = np.nan
    met = zephir[met_cols].copy()
    met[met>1000] = np.nan
    
    # Rotate the wind direction
    wd = wd + rot
    wd[wd>360] = wd[wd>360] - 360
    wd[wd<0] = wd[wd<0] +360
    
#    lat = []
#    lon = []
#    for i in zephir.index:
#        try:
#            lat = np.append(lat, float(zephir['GPS'].str.split(' ')[i][0]))
#            lon = np.append(lon, float(zephir['GPS'].str.split(' ')[i][1]))
#        except ValueError:
#            lat = np.append(lat, np.nan)
#            lon = np.append(lon, np.nan)
#    
#    position = pd.DataFrame({'lat':lat, 'lon':lon}, index = zephir.index)
    
    os.chdir(origin_dir)   
    
    return uh, wd, w, met#, position

def read_zephir_spectra(directory, filename):
    """
    Some test data:
    directory = '/scratch/Data/Zephir/Spectra/'
    filename = 'spectra_test.zph.csv'

    """
    
    spectra = pd.read_csv(directory+filename)
    
    # Seperate the Bins and get the Radial velocities
    bin_names = spectra.columns[spectra.columns.str.contains('FFT Bin ')]
    spectra_bins = np.array(spectra[bin_names])
    max_data = np.nanmax(spectra_bins,axis=1)
    
    new_spectra = spectra[spectra.columns[spectra.columns.str.contains('FFT Bin ')]].copy()
    new_spectra_index = pd.to_datetime(spectra['Time and Date'], format='%m/%d/%Y %I:%M:%S %p')
    
    bin_pos = []
    for i in range(len(spectra_bins)):
        bin_pos.append(np.where(spectra_bins[i,:]== max_data[i])[0][0])
    
    # Estimate the corresponding doppler frequency  
    doppler_freq = 195.31*np.array(bin_pos) *1e3
    
    # Estimate the radial velocities
    radial_vel_temp = 1.575 *1e-6 * doppler_freq/2.
    
    # Save the Radial velocities (corresponding to azimuth bin, height bin and time step)
    radial_vel = pd.DataFrame({'Timestamp':pd.to_datetime(spectra['Time and Date'], format='%m/%d/%Y %I:%M:%S %p'),
                               'Reference': spectra['Reference'],
                              'Height (m)': spectra['Height (m)'],
                              'Phase (rad)': spectra['Phase (rad)'],
                              'Radial velocities (m/s)':radial_vel_temp})
      
    ## Check the maximum estimate
#    plt.figure()
#    plt.pcolormesh(np.log(spectra[bin_names][100:150].transpose()))
#    plt.plot(np.array(bin_pos)[100:150],'r')  
    
    return radial_vel
      
# =============================================================================
# Windcube V2   
# =============================================================================

def get_WindcubeV2_header(filename):
    """
    Get the information of the Windcube V2 file header 
    """

    fil = open(filename,'rb')   
    lines = fil.readlines()
        
    header1 = lines[0].decode('latin-1')
    header1 = header1.replace('\r\n','')
    header1 = header1.split('=')
    HeaderLength = int(header1[1])
        
    header = [i.decode('latin-1') for i in lines[0:HeaderLength+1]]    
    header = [i.replace('\n','') for i in header]
    header = [i.replace('\t',' ') for i in header]
        
    # Build dictionary for the header 
    header_dic = {}
    for lin in header:
        if '=' in lin:
            splt = lin.split('=')           
            if splt[0] == 'InitialLD1(mA)' or splt[0] == 'InitialLD2(mA)':
                header_dic[splt[0]] = splt[2]
            elif len(splt)>2:
                print('manual add required, add -elif- statement')  
                print(splt)
            else:
                header_dic[splt[0]] = splt[1]   
                    
    # Get the column names from header
    column_names = lines[HeaderLength+1].decode('latin-1') 
    column_names = column_names.split('\t')                 # split where there was originally a tab (\t)
    column_names.remove('\n')                               # remove all \n (case for last enterence)
        
    # Get the height/altitude levels from header
    height_level = header_dic['Altitudes (m)'].split(' ')
    height_level = [int(i) for i in list(filter(None, height_level))]

    header     = header_dic 
    header_len = HeaderLength
    col_names  = column_names
    level      = height_level # Altitude level
    
    return header, header_len, col_names, level
        
def read_raw_WindcubeV2(filenames):
    """
    Read the raw data of the Windcube V2 dataset
    """
       
    for file in filenames:
            
        # get the header info
        header, header_len, col_names, level = get_WindcubeV2_header(file)
        buffer = pd.read_csv(filepath_or_buffer=file,
                             sep='\t',
                             header = None,
                             skiprows = header_len+2,
                             names = col_names,
                             usecols = range(len(col_names)),
                             index_col = False)
            
        # Convert index column to datetime object
        buffer.index = pd.to_datetime(buffer[col_names[0]].tolist(),
                                      format='%Y/%m/%d %H:%M:%S.%f')
        try:
            data
        except NameError:
            data = buffer
            
            
        # Accumulate read data in one DataFrame 
        data = pd.concat([data,buffer])  
        
    # Check if the right data location was defined
    try:
        data
    except UnboundLocalError:
        return print('No data found \nPlease enter a different data location in order to read raw data')

    # Remove possible duplicated rows
    data = data[~data.index.duplicated()]
    
    return data

def read_Windcube_processed_netcdf(directory,identifier):
    """
    """
    # Get the path of current (origin) directory
    origin_dir = os.getcwd()        
    # Change to directory that includes the files to be read                                      
    os.chdir(directory) 
      
    # Get filenames corresponding to ending .nc from current directory
    names = sorted(glob.glob('*'+identifier+'*.nc'))

    time  = {}
    W     = {} # vertical wind speed
    U   = {} # u wind speed
    V   = {} # v wind speed

    data_all = pd.DataFrame()      
    for name in names:
        dataset = Dataset(name, 'r')
        W[name] = dataset['z_wind_speed'][:,:].data
        W[name][W[name]<-10] = np.nan
        U[name] = dataset['x_wind_speed'][:,:].data
        U[name][U[name]<-90] = np.nan
        V[name] = dataset['y_wind_speed'][:,:].data
        V[name][V[name]<-90] = np.nan      
            
        time[name] = dataset['time'][:].data

        data = pd.DataFrame(index = pd.to_datetime(num2date(time[name],units = 'seconds since 1970-01-01 00:00:00')))
        for i in range(W[name].shape[1]):
            data['z_wind_speed_'+str(i)]     = W[name][:,i]    
            data['x_wind_speed_'+str(i)]     = U[name][:,i]
            data['y_wind_speed_'+str(i)]     = V[name][:,i]
                  
            # Estimate the horizontal wind speed and wind direction
            data['wind_speed_'+str(i)], data['wind_direction_'+str(i)] = get_wind_from_vector(u=U[name][:,i],v= V[name][:,i])
                  
        data_all = pd.concat([data_all,data],axis=0)
        dataset.close()
    os.chdir(origin_dir)      
    return data_all

# =============================================================================
# =============================================================================
# =============================================================================
########### Other remote sensing instrumentation ##############################
# =============================================================================
# =============================================================================
# =============================================================================

# =============================================================================
# Hatpro Radiometer
# =============================================================================

def read_Hatpro(directory,identifier):
    """
    """
    # Get the path of current (origin) directory
    origin_dir = os.getcwd()        
    # Change to directory that includes the files to be read                                      
    os.chdir(directory)     
    
    filenames = sorted(glob.glob('*'+identifier+'*.nc'))    
    
    data = pd.DataFrame()
    temp = pd.DataFrame()

    for file in filenames:
        buffer = pd.DataFrame()
        buffer_dict = {}
        dataset = Dataset(file, 'r')
        # Read the variables
        if identifier == 'iwv-lwp':
            buffer['liquid_water_path'] = dataset['liquid_water_path'][:].data
            buffer['integrated_water_vapour'] = dataset['integrated_water_vapour'][:].data
            buffer.index = num2date(dataset['time'][:].data, units = dataset['time'].units)  
            buffer['integrated_water_vapour'][buffer['integrated_water_vapour']<0] = np.nan
            buff_temp = pd.DataFrame()
        else:  
            for variable in dataset.variables.keys():
                if len(dataset[variable].shape) == 1:
                    buffer[variable] = dataset[variable][:]
                elif len(dataset[variable].shape) == 2:
                    buffer_dict[variable] = dataset[variable][:,:].data
                if 'time' in buffer.columns:
                    buffer.index = num2date(buffer.time, units = dataset['time'].units)  
            if identifier == 'boundary-layer-temperature-profiles':
                buff_temp = pd.DataFrame(buffer_dict['air_temperature'],
                                         index = buffer.index,
                                         columns = buffer_dict['altitude'][0,:])  
            elif identifier == 'moisture-profiles':
                 buff_temp = pd.DataFrame(buffer_dict['relative_humidity'],
                                          index = buffer.index,
                                         columns = buffer_dict['altitude'][0,:])    

        data = pd.concat([data,buffer], axis =0)
        temp = pd.concat([temp,buff_temp], axis =0)
        dataset.close()

    os.chdir(origin_dir)     
    return data, temp

def read_MRR(directory):
    """
    """
    origin_dir = os.getcwd()        
    # Change to directory that includes the files to be read                                      
    os.chdir(directory)     
    
    filenames = sorted(glob.glob('*.nc'))  

    v_t = pd.DataFrame()
    Ze  = pd.DataFrame()
    for file in filenames:
        dataset = Dataset(file)

        raw_time_buf = dataset['time'][:].data
        time = num2date(raw_time_buf, units = dataset['time'].units)
        spectral_vel_bin = dataset['velocity'][:].data
        height = dataset['height'][:].data

        # mean (integrated) velocity of most significant peak
        v_t_buf = dataset['W'][:].data
        v_t_buf[v_t_buf<=-9999] = np.nan
        v_t_buf = pd.DataFrame(v_t_buf, index=time, columns = height[0,:])
    
        # Get the equivalent Radar Reflectivity
        Ze_buf = dataset['Ze'][:].data
        Ze_buf[Ze_buf<=-9999] = np.nan
        Ze_buf = pd.DataFrame(Ze_buf, index=time, columns = height[0,:])

#        # Get the doppler spectral density
#        dsd = dataset['eta'][:].data

        dataset.close()
    
        # Combine the datasets
        v_t = pd.concat([v_t, v_t_buf], axis=0)
        Ze  = pd.concat([Ze, Ze_buf], axis=0)
    os.chdir(origin_dir)        
    return v_t, Ze

# =============================================================================
###############################################################################
################################ SATELLITE ####################################
###############################################################################
# =============================================================================
    
def read_sst_from_sat(filename, lon_lim = [-180,180], lat_lim = [-90,90]):
    """
    Some test data:
    directory = '/scratch/Data/SST_global_EC/'
    filename = '20180226120000-CMC-L4_GHRSST-SSTfnd-CMC0.1deg-GLOB-v02.0-fv03.0.nc'
    """
    
    # open the netcdf dataset
    dataset = Dataset(filename)

    # get all variable names from the dataset
    variables = [i for i in dataset.variables]
    
    # Get the dimension name of the dataset
    dim = [i for i in dataset.dimensions]    
    
    # Get the time vector of the dataset 
    time = num2date(dataset[dim[0]][:].data, units=dataset[dim[0]].units)    
    
    latitude  = dataset[variables[1]][:].data
    longitude = dataset[variables[2]][:].data
    
    sst       = dataset[variables[3]][0,:,:].data
    error     = dataset[variables[4]][0,:,:].data
    sea_ice   = dataset[variables[5]][0,:,:].data
    mask      = dataset[variables[6]][0,:,:].data
    
    # Define the longitude and longitude limitations
    lon_limits = np.where((longitude>lon_lim[0]) & (longitude < lon_lim[1]))[0]
    lat_limits = np.where((latitude>lat_lim[0]) & (latitude < lat_lim[1]))[0]
    
    dataset.close()
    # Return the part wanted for the analysis
    return time, latitude[lat_limits], longitude[lon_limits], \
           sst[:,lon_limits][lat_limits,:], sea_ice[:,lon_limits][lat_limits,:], \
           error[:,lon_limits][lat_limits,:], mask[:,lon_limits][lat_limits,:]
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Feb 24 15:05:09 2020

@author: cdu022

Toolbox for analysis and visualisation
Functions handy for analysis of (atmospheric) data 
"""

import pandas as pd
import numpy as np
import netCDF4 as nc
from netCDF4 import Dataset
from netCDF4 import num2date, date2num
import matplotlib.pyplot as plt
import cartopy.crs as ccrs
from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER
import matplotlib.ticker as mticker
from sklearn.linear_model import LinearRegression
from scipy import signal
from scipy.signal import butter, lfilter, freqz

    
# =============================================================================
# Retrieval Tool (Lidar observations)
# =============================================================================    

def DBS_least_squares_retrieval(all_beams, roll, pitch, yaw, az = [0,90,180,270,0], ze = [28,28,28,28,0], rot = False, order = 'zyx'):
    """
    find the 3 dimensional wind vector from radial velocity of a DBS scan, using the least squares method
    
    Ax = b      
    A = N*R 
    
    Least squares method:
    x = (A'A)^(-1)A'b
    
    N:     LOS transformation matrix
    R:     rotation matrix
    b:     radial velocity vector
    x:     estimate for 3D wind vector
    ':     transposed
    ^(-1): inversed
    
    Parameters
    ==========
    all_beams : pandas.DataFrame
        DataFrame containing all available beams of the DBS scan (should be interpolated in time, or filled with nearest)
        [B1, B2, B3, B4, B5]
    theta : float
        cone angle, θ, of DBS scan (rad)
    roll : array
       angle of rotation around x-axis of the system (rad or deg, set unit accordingly)
    pitch : array
        angle of rotation around y-axis of the system (rad or deg, set unit accordingly)
    yaw : array
        angle of rotation around z-axis of the system (rad or deg, set unit accordingly)
    rot : condition
        Indication, if motion compensation should be applied (True or False)
    order : string
        order in which the coordinate system should be rotated 
        xyz: Rotate around z first, than around the transformed y axis, at last around the twice transformed x axis
        zyx: Rotate around x first, than around the transformed y axis, at last around the twice transformed z axis
    """    
    # Define index vector (here corresponding to the time axis)
    time  = np.arange(len(all_beams.index))
    beams = len(all_beams.columns)
    
    # Convert radial velocity time series to 3D array   
    vr = np.array([[all_beams[beam].values] for beam in all_beams.columns])
    
    # Get the LOS transformation vector
    N  = get_N(time, az, ze, beams)    
    
    if rot:
        R = get_R(roll, pitch, yaw, time, order)
        A = multi_dim_matrix_mult(N,R,time)
    else:
        A = N
        
    # Transpose A
    At        = transpose_matrix(A)
    # Multiply Apt and Ap
    AtA       = multi_dim_matrix_mult(At,A,time)
    # find inverse of AptAp
    AtA_inv   = inverse_3x3_matrix(AtA)
      
    # Multiply AptAp_inv and Apt
    AtA_invAt = multi_dim_matrix_mult(AtA_inv,At,time)
    # Do the retrieval  
    retrieval = multi_dim_matrix_mult(AtA_invAt,vr,time)[:,0,:]
    # Convert the retrieved time series to a pandas DataFrame
    retrieval = pd.DataFrame(np.transpose(retrieval), columns = ['u','v','w'], index = all_beams.index)
    
    return retrieval    

# =============================================================================
# Required preparations of the radial velocity vector transformations
# =============================================================================

def get_N(time, az = [0,90,180,270,0], ze = [28,28,28,28,0], beams=5):
    """
    get the LOS transformation vector N
    
    vr = N*u
    
    N[i,:] = [sin(θ_i)*cos(α_i)  sin(θ_i)*sin(α_i)   cos(θ_i)]
    
    θ: zenith angle of the cone  (ze)
    α: azimuth angle of the cone (az)
    i: individual beam
    
    For the default setting (orienting on the WindCube V2) the LOS tansformation matrix will result:
    
    |vr1|   | sin(θ)  0    cos(θ)| 
    |vr2|   |   0   sin(θ) cos(θ)| |u|
    |vr3| = |-sin(θ)  0    cos(θ)|*|v|
    |vr4|   |   0  -sin(θ) cos(θ)| |w|
    |vr5|   |   0     0      1   |
    
    vr1: radial velocity along beam in opposite x-direction (B1)
    vr2: radial velocity along beam in opposite y-direction (B2)
    vr3: radial velocity along beam in x-direction          (B3)
    vr4: radial velocity along beam in y-direction          (B4)
    vr5: radial velocity along beam in z-direction          (B5)
    
    Parameters:
    ===========
    time : array
        1D array containing the time information of the series that is retrieve/analysed
    theta : float
        cone angle, θ, of DBS scan (rad)
    az : list
        azimuth angles corresponding to the beams (deg). Note! len(az) must be equal to beams.
    ze : list
        zenith angles corresponding to the beams (deg). Note! len(ze) must be equal to beams. 
    Returns:
    ========
    N : array
        the LOS transformation matrix [shape: beamsx3xlen(time)]
    """
    
    az = np.deg2rad(np.array(az))
    ze = np.deg2rad(np.array(ze))
    
    N = np.array([[np.cos(az[i])*np.sin(ze[i])*np.ones(len(time)), 
                   np.sin(az[i])*np.sin(ze[i])*np.ones(len(time)), 
                   np.cos(ze[i])*np.ones(len(time))] for i in range(beams)])    

    return N
    
def get_R(roll, pitch, yaw, time, order = 'xyz', unit = 'deg'):
    """
    Get the rotation Matrix
    
    Rx : rotation matrix around the x-axis
    Ry : rotation matrix around the y-axis
    Rz : rotation matrix around the z-axis
    
    Parameters:
    ===========
    roll : array
       angle of rotation around x-axis of the system (rad or deg, set unit accordingly)
    pitch : array
        angle of rotation around y-axis of the system (rad or deg, set unit accordingly)
    yaw : array
        angle of rotation around z-axis of the system (rad or deg, set unit accordingly)
    time : array
        1D array containing the time information of the series that is retrieve/analysed
    order : string
        order in which the coordinate system should be rotated 
        xyz --> Rotate around z first, than around the transformed y axis, at last around the twice transformed x axis
        zyx --> Rotate around x first, than around the transformed y axis, at last around the twice transformed z axis
    """
    if unit == 'deg':
        roll  = np.deg2rad(roll)
        pitch = np.deg2rad(pitch)
        yaw   = np.deg2rad(yaw)
    
    # Define the rotational matrices
    Rx = np.array([[np.ones(len(roll)), np.zeros(len(roll)),np.zeros(len(roll))],
                   [np.zeros(len(roll)), np.cos(roll), -np.sin(roll)],
                   [np.zeros(len(roll)), np.sin(roll),np.cos(roll)]])

    Ry = np.array([[np.cos(pitch), np.zeros(len(pitch)), np.sin(pitch)],
                   [np.zeros(len(pitch)),np.ones(len(pitch)),np.zeros(len(pitch))],
                   [-np.sin(pitch), np.zeros(len(pitch)), np.cos(pitch)]])
    
    Rz = np.array([[np.cos(yaw), -np.sin(yaw), np.zeros(len(yaw))],
                   [np.sin(yaw), np.cos(yaw), np.zeros(len(yaw))],
                   [np.zeros(len(yaw)),np.zeros(len(yaw)),np.ones(len(yaw))]])

    # Usually R should be rotated: 
    Rx = transpose_matrix(Rx)
    Ry = transpose_matrix(Ry)
    Rz = transpose_matrix(Rz)
    
    # Get the rotation matrix including rotation around x, y and z axis
    if order == 'xyz':
        # Rotate around x first, than around the transformed y axis
        RyRx   = multi_dim_matrix_mult(matrix1=Ry,matrix2=Rx,time=time)
        # Additionally rotate around z
        R      = multi_dim_matrix_mult(matrix1=Rz,matrix2=RyRx,time=time) 
    elif order == 'zyx':
        # Rotate around z first, than around the transformed y axis
        RyRz   = multi_dim_matrix_mult(matrix1=Ry,matrix2=Rz,time=time)
        # Additionally rotate around x
        R      = multi_dim_matrix_mult(matrix1=Rx,matrix2=RyRz,time=time)     

    return R

def correct_LOS_vel(all_beams, u_plat, az = [0,90,180,270,0], ze = [28,28,28,28,0]):
    """
    Function to remove translational velocity of a platform from the radial velocity vector 
    
    vr_plat = N*u_plat
    vr_corr = vr-vr_plat
    
    u_plat:  velocity of the platform (kartesian grid) 
    vr_plat: radial velocity of the platform, transposed to LOS  
    N:       the LOS transformation matrix
    
    Parameters
    ==========
    all_beams : pandas.DataFrame
        DataFrame containing all available beams of the DBS scan (should be interpolated in time, or filled with nearest)
        [B1, B2, B3, B4, B5]
    u_plat : array
        velocity (u,v,w) of the platform [size: 3x1xlen(time)]
    theta : float
        cone angle, θ, of DBS scan (rad)
        
    Returns:
    ========
    corr_beams : DataFrame
        radial velocity corrected by platform velocity [shape: len(beams) x len(time)]
    """
    
    time = all_beams.index
    
    # Get the LOS transformation vector
    N  = get_N(time, az, ze, beams = len(all_beams.columns))    
    
    # Estimate the radial velocity vector of the platform
    vr_plat = multi_dim_matrix_mult(N,u_plat,time)[:,0,:]
    
    # convert to DataFrame
    vr_plat = pd.DataFrame(np.transpose(vr_plat), 
                           index = time, 
                           columns = all_beams.columns)
    # Get the radial velocity, corrected by the LOS transformed platform velocity
    corr_beams = all_beams.copy() - vr_plat.copy()
    
    return corr_beams

# =============================================================================
# Required matrix operations for least squares retrieval
# =============================================================================   

def transpose_matrix(matrix):
    """
    method to transpose a time series of matrices
    
    A(t) --> A'(t)
    
    Parameters:
    ===========
    matrix : array
        time series of matrices to be transposed 
        
    transpose : array
        time series of transposed matrices
    """
    transpose = np.zeros((matrix.shape[1],matrix.shape[0],matrix.shape[2]))
    for i in range(matrix.shape[0]):
        for j in range(matrix.shape[1]):
            transpose[j,i,:] =  matrix[i,j,:]
    return transpose

def multi_dim_matrix_mult(matrix1,matrix2,time):
    """
    function that is capable of multiplying time series of matrices
    
    Parameters
    ----------
    matrix1 : array
        left hand side 3D matrix
    matrix2 : array
        right hand side 3D matrix
    Returns
    -------
    res : array
        3D matrix product (time series of multiplied matrices)
    
    """
    res = [[[0 for z in range(len(time))] for x in range(len(matrix2[0,:]))] for y in range(len(matrix1[:,0]))]  
         
    # explicit for loops 
    for i in range(len(matrix1[:,0])): 
        for j in range(len(matrix2[0,:])): 
            for k in range(len(matrix2[:,0])): 
         
                # resulted matrix 
                res[i][j][:] += matrix1[i][k][:] * matrix2[k][j][:]
    return np.array(res)

def inverse_3x3_matrix(matrix):
    """
    method to estimate the inverse of a time series of 3x3 matrices
    
    A(t) --> A^(-1)(t)
    
    Parameters:
    ===========
    matrix : array
        time series of matrices to be inverted 
        
    inv_matrix : array
        time series of inverse matrices
    """
    inv_matrix = np.zeros((matrix.shape[0],matrix.shape[1],matrix.shape[2]))
       
    # Find the determinante
    det = matrix[0,0,:]*matrix[1,1,:]*matrix[2,2,:] \
          +matrix[0,1,:]*matrix[1,2,:]*matrix[2,0,:] \
          +matrix[0,2,:]*matrix[1,0,:]*matrix[2,1,:] \
          -matrix[0,2,:]*matrix[1,1,:]*matrix[2,0,:] \
          -matrix[0,0,:]*matrix[1,2,:]*matrix[2,1,:] \
          -matrix[0,1,:]*matrix[1,0,:]*matrix[2,2,:]
    # Get the inverse of the determinante
    inv_det = 1./det
    
    # get the inverse matrix
    inv_matrix[0,0,:] = inv_det*(matrix[1,1,:]*matrix[2,2,:]-matrix[2,1,:]*matrix[1,2,:])
    inv_matrix[0,1,:] = inv_det*(matrix[0,2,:]*matrix[2,1,:]-matrix[0,1,:]*matrix[2,2,:])
    inv_matrix[0,2,:] = inv_det*(matrix[0,1,:]*matrix[1,2,:]-matrix[0,2,:]*matrix[1,1,:])
    inv_matrix[1,0,:] = inv_det*(matrix[1,2,:]*matrix[2,0,:]-matrix[2,2,:]*matrix[1,0,:])
    inv_matrix[1,1,:] = inv_det*(matrix[0,0,:]*matrix[2,2,:]-matrix[0,2,:]*matrix[2,0,:])
    inv_matrix[1,2,:] = inv_det*(matrix[0,2,:]*matrix[1,0,:]-matrix[0,0,:]*matrix[1,2,:])
    inv_matrix[2,0,:] = inv_det*(matrix[1,0,:]*matrix[2,1,:]-matrix[1,1,:]*matrix[2,0,:])
    inv_matrix[2,1,:] = inv_det*(matrix[0,1,:]*matrix[2,0,:]-matrix[0,0,:]*matrix[2,1,:])
    inv_matrix[2,2,:] = inv_det*(matrix[0,0,:]*matrix[1,1,:]-matrix[0,1,:]*matrix[1,0,:])
    return inv_matrix    

# =============================================================================
# Statistical Tools
# =============================================================================

def fft(data, dt):
    """
    Estimate the Energy spectrum (depending on frequency or wavenumber) 
    from fourier transform of time ore spacial series 
    """
    data_mask = np.arange(len(data))
    mask = np.isnan(data)
      
    if len(np.where(np.isnan(data))[0]) == len(data):
        N  = len(data)
        T  = dt*N
        df = 1.0/T
        f = np.fft.rfftfreq(N)*N*df
            
        E = pd.DataFrame(f[1:int(N/2)]*np.nan, index = f[1:int(N/2)])
    else:      
        xfiltered = np.interp(data_mask, data_mask[~mask], data[~mask])
            
        N  = len(xfiltered)
        T  = dt*N
            
        df = 1.0/T      # Maximum Frequency
        dw = 2.*np.pi*df
        ny = dw*N/2
            
        a = np.fft.rfft(xfiltered)
        f = np.fft.rfftfreq(N)*N*df
            
        # get the energy content corresponding to each of the frequencies
        E = pd.DataFrame(abs(a[1:int(N/2)]*np.conj(a[1:int(N/2)])), index = f[1:int(N/2)])

    return E


def do_regression(x,y):
    """
    Do a linear regression between two Datasets x & y
    """
    # Remove all points corresponding to nan values
    coords = np.where((~np.isnan(x)) & (~np.isnan(y)))[0]

    x = x[coords].reshape(-1, 1)
    y = y[coords].reshape(-1, 1)
    
    # Utilize the linear regression model
    reg = LinearRegression().fit(x, y)
#    reg_coef  = reg.coef_[0,0]
#    reg_score = reg.score(x, y)
    return x,y,reg
    
def cal_consistency_param(x_array, y_array, condition):
    """
    Do a full consistency/ comparison analysis corresponding to a range of conditions
    """
    reg_coef  = np.zeros(len(condition))
    reg_score = np.zeros(len(condition))
    bias      = np.zeros(len(condition))
    sde       = np.zeros(len(condition))
    corr_coef = np.zeros(len(condition))
    c = 0
    for i in condition:
        x = x_array[i].values
        y = y_array[i].values
        # estimate the bias, sde & correlation
        bias[c]      = np.nanmean(y-x)
        sde[c]       = np.nanstd(y-x)
        corr_coef[c] = np.corrcoef(x,y)[0,1]
    
        x,y,reg = do_regression(x,y)
    
        reg_coef[c]  = reg.coef_[0,0]
        reg_score[c] = reg.score(x, y) # Mearsure for correlation?

        c = c+1    
    return bias, sde, reg_coef, reg_score, corr_coef

# =============================================================================
# Filtering tools
# =============================================================================

def butter_lowpass(cutoff, fs, order=5):
    nyq = 0.5 * fs
    normal_cutoff = cutoff / nyq
    b, a = butter(order, normal_cutoff, btype='low', analog=False)
    return b, a


def butter_lowpass_filter(data, cutoff, fs, order=5):
    b, a = butter_lowpass(cutoff, fs, order=order)
    y = lfilter(b, a, data)
    return y

def butter_highpass(cutoff, fs, order=5):
    nyq = 0.5 * fs
    normal_cutoff = cutoff / nyq
    b, a = signal.butter(order, normal_cutoff, btype='high', analog=False)
    return b, a

def butter_highpass_filter(data, cutoff, fs, order=5):
    b, a = butter_highpass(cutoff, fs, order=order)
    y = signal.filtfilt(b, a, data)
    return y


# =============================================================================
# Wind analysis tools    
# =============================================================================

def get_wind_vector(u_h,wd):
    """
    Estimate the horizontal wind speed components u & v from horizontal wind speed and direction
    Transformation from polar coodinates to cartesian coordinates
    
    Parameters
    ----------
    u_h : float
        horizontal wind speed
    wd : float
        horizontal wind direction
    
    Returns
    -------
    u : float
        longitudinal wind component
    v : float
        latitudinal wind component
    """
   
    # transform wd unit (deg) to radians
    wd_rad = wd/180*np.pi
    
    u = - u_h * np.sin(wd_rad)
    v = - u_h * np.cos(wd_rad)
    
    return u,v

def get_wind_from_vector(u,v):
    """
    Estimate the horizontal wind speed and direction from the horizontal wind speed components u & v
    Transformation from cartesian coordinates to coodinates
    
    Parameters
    ----------
    u : float
        array of longitudinal wind component    
    v : float
        array of latitudinal wind component
    Results
    -------
    u_h : float
        array of horizontal wind speed
    wd : float
        array of horizontal wind direction        
    """
    
    # Get the horizontal wind speed
    u_h = np.sqrt(u**2+v**2)
    
    # Get the wind direction
    wd = 180/np.pi * np.arctan(u/v)
    
    # Differentiate between the different wind direction sectors
    wd[(u>=0)&(v>=0)] = wd[(u>=0)&(v>=0)] + 180
    wd[(u<=0)&(v>=0)] = wd[(u<=0)&(v>=0)] + 180     
    wd[(u>=0)&(v<=0)] = wd[(u>=0)&(v<=0)] + 360    
    # Make sure no wind direction value exeeds 360 or falls below 0
    wd[wd>=360] = wd[wd>=360] - 360
    wd[wd<   0] = wd[wd<   0] + 360
    
    return u_h, wd

def rot_wind_vector_in_dir(U,V):
    """
    Function to estimate the along wind or parallel to mean wd (new_U) and 
    cross wind or perpendicular to mean wd (new_V) components 
     
    Parameters
    ----------
    u : float
        pandas Dataframe (with datetime index) of longitudinal wind component (in east west direction)
    v : float
        pandas Dataframe (with datetime index) of latitudinal wind component (in north south direction)

    Results
    -------
    new_U : float
        pandas Dataframe (with datetime index) of along wind component or parallel to mean wd
    new_V : float
        pandas Dataframe (with datetime index) of cross wind component or perpendicular to mean wd
    """
    mean_U = U.rolling(600,0,center=True).mean()#.resample('10min',loffset='5min').mean().copy()
    mean_V = V.rolling(600,0,center=True).mean()#.resample('10min',loffset='5min').mean().copy()
      
    mean_wind, mean_wd = get_wind_from_vector(mean_U,mean_V)
    wd_ups = 270.-mean_wd#.reindex(U.index, method='nearest') 
      
    new_U = U*np.cos(wd_ups/180.*np.pi) +V*np.sin(wd_ups/180.*np.pi)
    new_V = -U*np.sin(wd_ups/180.*np.pi) + V*np.cos(wd_ups/180.*np.pi)
      
    return new_U, new_V

def alt_rot_wind_vector_in_dir(U,V):
    """
    alternative rotate wind in mean wind direction
    """
    mean_U = U.rolling(600,0,center=True).mean()#.resample('10min',loffset='5min').mean().copy()
    mean_V = V.rolling(600,0,center=True).mean()#.resample('10min',loffset='5min').mean().copy()
      
    mean_wind, mean_wd = get_wind_from_vector(mean_U,mean_V)
    wind, wd = get_wind_from_vector(U,V)
    
    # Get wind direction fluctuations
    del_wd = (mean_wd-wd)*np.pi/180.
    
    new_U = wind*np.cos(del_wd)
    new_V = wind*np.sin(del_wd)

# =============================================================================
# Stability (conversion) Tools
# =============================================================================

def calc_p_z(p0,T0,z):
    """
    Calculate the pressure at a certain altitude
    
    Parameters
    ==========
    p0 : float
        Pressure at sea level
    T0 : float
        Temperature at sea level
    z : float
       altitude above sea level
    """
    # Earth-surface gravitational acceleration
    g = 9.81
    # Constant-pressure specific heat
    cp = 1004.68506
    # Molar mass of dry air
    M = 0.02896968
    # Universal gas constant
    R0 = 8.314462618
    
    return p0*(1 - g/(cp*T0)*z)**(cp*M/R0)

def cal_e_sat(T):
    """
    Calculate the water vapour saturation pressure
    
    Parameters
    ==========
    T : float
        Air Temperature at the Height of interest [K]
    """
    
    A = 6.112 
    m = 17.62
    Tn = 243.12
    T = T-273.15 # Convert to deg C
    
    Pws = A * np.exp((m*T)/(Tn+T))
    
    return Pws

def cal_e(T,RH):
    """
    Calculate the water vapour pressure
      
    Parameters
    ==========      
    Pws : water vapour saturation pressure (Is dependent on the temperature)
    """
    Pws = cal_e_sat(T)
    
    return Pws*RH/100.

def cal_mix_ratio(P,T,RH):
    """
    Calculate the mixing ratio
    """    
    Pw = cal_e(T,RH)
    return 0.622 * Pw/(P-Pw)

def cal_Td(P,T,RH):
    """
    calculate dew point
    """
    
    r = cal_mix_ratio(P,T,RH)
    
    A = 6.112 
    m = 17.62
    Tn = 243.12
    
    return Tn*np.log(r*P/((0.622+r)*A))/(m-np.log(r*P/((0.622+r)*A)))+273.15

def cal_virt_T(P,T,RH):
    """
    Calculate the virtual temperature
    """
    e = cal_e(T,RH)    
    q_v = 0.622*e/P    

    T_v = T*(1+0.608*q_v)
    
    return T_v 

def cal_pot_T(P,T,RH):
    """
    Calculate potential and virtual potential temperature
    """
#    r = cal_mix_ratio(P,T,RH)
    
    # Estimate virtual temperature
    T_v = cal_virt_T(P,T,RH)
    
    # Potential Temperature
    theta = T*(1000./P)**0.286
    
    # Virtual potential temperature    
    theta_v = T_v*(1000./P)**0.286 #theta*(1+0.61*r)
    
    return theta, theta_v

def cal_N(theta,alt):
    """
    Calculate the Brunt-Väisälä frequency
    
    """      
    g = 9.81

    del_thetha = theta[:,:-1]-theta[:,1:]
    del_z      = alt[:,:-1] - alt[:,1:]

    mean_theta = (theta[:,:-1]+theta[:,1:])/2
    mean_alt = (alt[:,:-1] + alt[:,1:])/2
    
    N = g/mean_theta * del_thetha/del_z
    
    return N, mean_alt

def cal_M(U,V,alt):
    """
    Calculate the shear frequency
    """
    del_U = U[:,:-1] - U[:,1:]
    del_V = V[:,:-1] - V[:,1:]
    del_z      = alt[:,:-1] - alt[:,1:]
    
    M = (del_U/del_z)**2 + (del_V/del_z)**2

    return M

def cal_Ri_number(P,T,RH, U_diff, V_diff, z):
    """
    Calculate the gradient Richardson number
    """
    # Get the gravity constant
    g   = 9.81 # kg/m^2s
    
    theta, theta_v = cal_pot_T(P,T,RH)
    
    # Estimate Brunt-Vaisaila Frequency (Buoyancy)
    N2 = (g/theta_v * theta_v.diff()/z.diff())
    # Estimate shear frequency
    M2 = (U_diff/z.diff())**2+(V_diff/z.diff())**2
    Ri = N2/M2
    
    return Ri

def cal_moist_lapse_rate(P,T,RH):
    """
    Calculate the moist adiabatic lapse rate
    """
    # Define constants
    g   = 9.8076  # m/s**2
    Hv  = 2501000 # J/kg
    Rsd = 287     # J/kgK
    Rsw = 461.5   # J/kg
#    eps = 0.622   #
    cpd = 1003.5  # J/kgK
    
    # estimate other needed parameters
    r = cal_mix_ratio(P,T,RH)
    
    # estimate moist adiabatic lapse rate (dependent on T)
    return g * (1+ (Hv*r/(Rsd*T)))/(cpd+Hv**2*r/(Rsw*T**2))*1000

# =============================================================================
# Profile analysis --- mainly for temperature, humidity and wind
# =============================================================================

def get_adiabates():
    """
    Estimate the dry and moist adiabate for a certain temperature range (Both Celsius and Kelvin)
    """
    # Define essential parameters
    R=287.04  #Jkg^-1K^-1
    cp=1005   #Jkg^-1K^-1
    p0=1000 #hPa
    L=2.5*10**6 #J kg^-1

    temp_celsius = np.array(range(-80,41))
    temp_kelvin=temp_celsius+273.15
    pressure_hPa=np.array(range(1050,90,-10))

    #make a grid of the data
    tempdata,pressdata=np.meshgrid(temp_kelvin,pressure_hPa*100.)

    #Initialise the arrays
    pot_temp_kelvin=np.zeros(tempdata.shape)
    es=pot_temp_kelvin
    ms=es
    pseudo_pot_temp_kelvin=es

    #Get the potential temperature                         
    pot_temp_kelvin=tempdata*(p0*100/pressdata)**(R/cp)

    #Get the saturation mix ratio
    #first the saturation vapor pressure after Magnus
    #Definition of constants for the Magnus-formula
    c1=17.62
    c2=243.12

    es=6.112*np.exp((17.62*(tempdata-273.15))/(243.12+tempdata-273.15)) #hPa
    #Now I'm calculation the saturation mixing ratio
    ms=622*(es/(pressdata/100-es)) #g/kg

    #At least I need the pseudo-adiabatic potential temperature
    pseudo_pot_temp_kelvin=pot_temp_kelvin*np.exp(L*ms/1000./cp/tempdata)

    #define the levels which should be plotted in the figure
    levels_theta=np.array(range(200,405,5))
    levels_ms=np.array([0.1,0.2,0.5, 1.0, 1.5, 2.0, 3.0, 4.0, 6.0, 8.0, 10.0, 12.0, 15.0, 20.0, 25.0, 30.0])
    levels_theta_e=np.array(range(220,410,10))

    #The plot
    fig = plt.figure(figsize=(10,10))
    theta=plt.contour(temp_celsius,pressure_hPa,pot_temp_kelvin,levels_theta,colors='blue')
    #plt.clabel(theta,levels_theta[0::2], fontsize=10, fmt='%1.0f')

    #sat_mix_ratio=plt.contour(temp_celsius,pressure_hPa,ms,levels_ms,colors='green')
    #plt.clabel(sat_mix_ratio,fontsize=10,fmt='%1.1f')
    theta_e=plt.contour(temp_celsius,pressure_hPa,pseudo_pot_temp_kelvin,levels_theta_e,colors='red')
    plt.clabel(theta_e,levels_theta_e[1::2],fontsize=10,fmt='%1.0f')

    plt.xlabel('Temperature [$^\circ$C]')
    plt.ylabel('Pressure [hPa]')
    plt.xticks(range(-80,45,5))
    plt.xlim((-80,50))
    plt.yticks(range(1050,50,-50))
    plt.gca().invert_yaxis()
    plt.grid(color='black',linestyle='-')

    # Get the moist adiabates
    moist_adiabates = pd.DataFrame()
    dry_adiabates = pd.DataFrame()
    # Get the moist adiabates
    for i in range(len(theta_e.allsegs)):
        buffer = pd.DataFrame(theta_e.allsegs[i][0][:,0], index = theta_e.allsegs[i][0][:,1], columns =[i])
        moist_adiabates = pd.concat([moist_adiabates, buffer], axis=0)
        buffer = pd.DataFrame(theta.allsegs[i][0][:,0], index = theta.allsegs[i][0][:,1], columns =[i])
        dry_adiabates = pd.concat([dry_adiabates, buffer], axis=0)
    
    moist_adiabates = moist_adiabates.sort_index().interpolate(method='pchip')+273.15
    dry_adiabates = dry_adiabates.sort_index().interpolate(method='pchip')+273.15

    return dry_adiabates, moist_adiabates

# =============================================================================
# Plotting tools
# =============================================================================

def make_map(projection=ccrs.PlateCarree()):
    fig, ax = plt.subplots(figsize=(9, 13),
                           subplot_kw=dict(projection=projection))
    gl = ax.gridlines(draw_labels=True)
    gl.xlabels_top = gl.ylabels_right = False
    gl.xformatter = LONGITUDE_FORMATTER
    gl.yformatter = LATITUDE_FORMATTER
    return fig, ax


# =============================================================================
# Tools to save data in a certain format 
# =============================================================================

def save_to_netcdf(global_atrributes, dimensions, variables, path, filename):
    """
    Save a dataset to netcdf format
          
    Parameters
    ----------
          
    global_atrributes : dict
        dictionary containing all global attributes that will be displayed in the netcdf file
          
    dimensions : dict
        dictionary containing all dimensions that are used in the netcdf file
              
          
    Returns
    -------
    """          
    # Open the netcdf file (in write)          
    dataset = Dataset(path + filename, 'w', format = 'NETCDF4_CLASSIC')
    # Set global attributes
    for attr in global_atrributes.keys():
        dataset.setncattr(attr, global_atrributes[attr])
      
    # Define Set of dimensions used for the variables
    # dataset includes time, height, latitude and longitude information (profile trajectory)
    if 'level' in dimensions.keys():
        level      = dataset.createDimension('level', dimensions['level'])
    time       = dataset.createDimension('time', dimensions['time'])
    latitude   = dataset.createDimension('latitude', dimensions['latitude'])
    longitude  = dataset.createDimension('longitude',dimensions['longitude'])
          
    # Define the Variables to be used
    for var in variables.keys():

        if len(variables[var]['dimension']) == 2:
            dataset.createVariable(var, variables[var]['precision'], 
                               (variables[var]['dimension'][0],
                                variables[var]['dimension'][1]), 
                                fill_value=-9999)            
            # write data to variables
            dataset.variables[var][:,:] = variables[var]['data']
        else:
            dataset.createVariable(var, variables[var]['precision'], 
                                   (variables[var]['dimension'],), fill_value=-9999) 
            # write data to variables
            dataset.variables[var][:] = variables[var]['data']
                      
        # set long names      
        dataset.variables[var].long_name = variables[var]['long_name']
        # set units 
        dataset.variables[var].units = variables[var]['units']
    
    # close the dataset
    dataset.close()


def discrete_cmap(N, base_cmap=None):
    """Create an N-bin discrete colormap from the specified input map"""

    # Note that if base_cmap is a string or None, you can simply do
    #    return plt.cm.get_cmap(base_cmap, N)
    # The following works for string, None, or a colormap instance:

    base = plt.cm.get_cmap(base_cmap)
    color_list = base(np.linspace(0, 1, N))
    cmap_name = base.name + str(N)
    return base.from_list(cmap_name, color_list, N)
                    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Feb 20 14:54:11 2020

@author: cdu022

The processing skript for the OBLO infrastructure
"""

from roblo import *
from toblo import *
import pandas as pd
import numpy as np
import glob


# =============================================================================
# Processing Methods (Instruments)
# =============================================================================

def process_radial_vel_zephir(directory, filename):
    """
    Get the Wind speed and direction from Radial velocities
    """
    
    radial_vel = read_zephir_spectra(directory, filename)
    
    # Seperate the series for each scan (same reference)
    first_scan = np.array(radial_vel['Reference'])[0]
    last_scan  = np.array(radial_vel['Reference'])[-1]
    
    h_buf = []
    U_buf = []
    w_buf = []
    wd_buf = []
    h_buf = []    
    for i in np.arange(first_scan,last_scan+1):
        if len(radial_vel[radial_vel['Reference']==i]) > 0:
            # Radial velocities corresponding to one scan
            vr_vec = radial_vel['Radial velocities (m/s)'][radial_vel['Reference']==i]
            # azimuth/ phase corresponding to one scan
            phase_vec = radial_vel['Phase (rad)'][radial_vel['Reference']==i]
            # get the height of scan
            height = np.median(radial_vel['Height (m)'][radial_vel['Reference']==i])
        
            # Get the "wind direction" from maximal correlation
            offset = np.arange(0,np.pi+0.1,0.1) # Define the offset
            corr = np.zeros(len(offset))
            for j in np.arange(len(offset)):
                corr[j] = np.corrcoef(np.array(0.5*abs(np.cos(phase_vec-offset[j]))),
                                      np.array(vr_vec))[0,1]

            wd_apx = offset[np.where(corr==np.max(corr))[0][0]]

            # Do a least square fit for the absolute wind vector components: ¦Uh¦ and ¦w¦ 
            # Define the radial velocity matrix
            b = np.transpose(np.matrix(vr_vec))
            ### use the 2D approach ###
            At = np.matrix([np.array(0.5*abs(np.cos(phase_vec-wd_apx))),
                            np.cos(1./6.*np.pi)*np.ones(len(phase_vec))])
            # If u,v and w component is to be retrieved
            #At = np.matrix([np.array(0.5*abs(np.cos(phase_vec))),np.array(0.5*abs(np.sin(phase_vec))),np.cos(1./6.*np.pi)*np.ones(len(phase_vec))])

            # Transpose A
            A = np.transpose(At)
            # Multiply At and A
            [U,w] = np.linalg.inv(At*A)*At*b
            U = U[0,0]
            w = w[0,0]

            h_buf  = np.append(h_buf, height)
            U_buf  = np.append(U_buf, U)
            w_buf  = np.append(w_buf, w)
            wd_buf = np.append(wd_buf, np.round(180/np.pi*wd_apx))

            retrieval = pd.DataFrame({'height':h_buf,
                                      'U':U_buf,
                                      'w':w_buf,
                                      'WD':wd_buf})

#            plt.figure()
#            plt.title('h = '+str(int(height))+' U = '+str(np.round(U,1))+' w = '+str(np.round(w,2))+r' $\theta$ = '+str(np.round(180/np.pi*wd_apx)))
#            plt.plot(vr_vec)
#            plt.plot(abs(0.5*U*np.cos(phase_vec-wd_apx)+w*np.cos(1./6.*np.pi)))
      
    # Sort retrieval by height
    levels = [10, 20, 30, 50, 60, 80, 100, 125, 150, 200] 
    U_sorted = {}
    w_sorted = {}
    wd_sorted = {}
    
    for level in levels:
        U_sorted[level] = np.array(retrieval['U'][retrieval['height']==level])
        w_sorted[level] = np.array(retrieval['w'][retrieval['height']==level])
        wd_sorted[level] = np.array(retrieval['WD'][retrieval['height']==level])
    # estimate u and v
    u_sorted, v_sorted = get_wind_vector(u_h=U_sorted,wd=wd_sorted)
    
    #plt.figure(figsize=[5,5])
    #plt.scatter(u_sorted.mean(),v_sorted.mean())    
    #plt.xlim([-20,20])
    #plt.ylim([-20,20])
    
    return u_sorted, v_sorted, w_sorted, U_sorted, wd_sorted


def process_ship_based_zephir(path_zephir, path_ship, delimiter,rot = 0, levels = [10, 20, 30, 38, 50, 60, 80, 100, 125, 150, 200]):
    """
    correct for ship movement and yawing of ship
    
    path_zephir = '/scratch/Data/GEOF232_2020/Zephir/Wind_Data/'
    delimiter = ';'
    path_ship = '/scratch/Data/GEOF232_2020/Ship_log/raw/'
    """
    
    # =============================================================================
    # Read the processed data
    # =============================================================================

    uh, wd, w, met = read_zephir(path = path_zephir, 
                                           delimiter=delimiter, 
                                           rot = rot)

    uh.columns = levels
    wd.columns = levels

    # Convert data to u and v components
    u, v = get_wind_vector(u_h = uh , wd = wd)

    # =============================================================================
    # Read ship log data (Kristine Bonnevie)
    # =============================================================================
    files = glob.glob(path_ship+'*.csv')
    log = pd.DataFrame()
    for file in files:
        print(file)
        log_buf = read_kristine_bonnevie(file)
        log = pd.concat([log,log_buf], axis=0)
    log = log.sort_index()
    
    # Get the positions
    lat = log['Latitude'][~log.index.duplicated()].reindex(u.index, method='nearest')
    lon = log['Longitude'][~log.index.duplicated()].reindex(u.index, method='nearest')
    
    position = pd.concat([lon,lat], axis=1)
    
    # Get SST and AirT
    airT = log['Air temp'][~log.index.duplicated()].reindex(u.index, method='nearest')
    ssT  = log['Water temp'][~log.index.duplicated()].reindex(u.index, method='nearest') 
    
    # Get Air Pressure 
    p = log['Air pressure'][~log.index.duplicated()].reindex(u.index, method='nearest')
    
    # Get humidity
    rh = log['Humidity'][~log.index.duplicated()].reindex(u.index, method='nearest')
    
    # Define yaw/heading
    yaw = (log['Heading'][~log.index.duplicated()].reindex(u.index, method='nearest'))/180*np.pi
    
    # Get the ship speed 
    ship_speed = (log['Speed']*1.8)/3.6 # convert from knots to m/s
    ship_speed = ship_speed[~log.index.duplicated()].reindex(uh.index, method = 'nearest')

    # Get u and v component of ship speed
    u_s =   ship_speed * np.cos(yaw) 
    v_s = - ship_speed * np.sin(yaw)
    

    # and of wind speed collected by ship
    us, vs = get_wind_vector(u_h =  log['Wind'], wd = log['Wind dir'])
    
    # Convert to same time resolution as Lidar observations
    us_new = us[~us.index.duplicated()].reindex(u.index, method = 'nearest').copy()
    vs_new = vs[~vs.index.duplicated()].reindex(u.index, method = 'nearest').copy()
    
    # Get the corresponding wind speed and direction (for this resolution)
    ush_new, wds_new = get_wind_from_vector(u = us_new, v = vs_new)

    # =============================================================================
    # Correct the wind data
    # =============================================================================

    # rotate the wind components 
    u_rot = {}
    v_rot = {}
    for level in levels:
        u_rot[level] =  u[level]*np.cos(yaw)+v[level]*np.sin(yaw)
        v_rot[level] = -u[level]*np.sin(yaw)+v[level]*np.cos(yaw)
    u_rot = pd.DataFrame(u_rot)
    v_rot = pd.DataFrame(v_rot)    

    # Get the wind speed and direction from the rotated coordinates
    uh_new, wd_new = get_wind_from_vector(u = u_rot, v = v_rot)

    # Correct for rotated wd estimates of Lidar
    for level in levels:
        wd_buf = wd_new[level].copy()
        wd_buf = pd.DataFrame(wd_buf, columns = [level])
        wd_buf[abs(wd_buf[level] - wds_new)>135] = wd_buf[abs(wd_buf[level] - wds_new)>135] -180
        wd_buf[wd_buf<0] = wd_buf[wd_buf<0] +360
        wd_buf[wd_buf>360] = wd_buf[wd_buf>360]-360
        
        wd_new[level] = wd_buf.copy()
        
    # Get the u and v components from updated wind direction
    u_new, v_new = get_wind_vector(u_h =  uh, wd = wd_new)
    
    # Get the final wind components (corrected for ship movement)
    u_f = {}
    v_f = {} 
    for level in levels:
        u_f[level] = u_new[level] + u_s
        v_f[level] = v_new[level] + v_s
    u_f = pd.DataFrame(u_f)
    v_f = pd.DataFrame(v_f)
    
    uh_f, wd_f = get_wind_from_vector(u = u_f, v = v_f)
    
    return u_f, v_f, uh_f, wd_f, ush_new, wds_new, ship_speed, yaw, position, airT, ssT, p, rh

def seperate_processed_Windcube_components(lidar_data):
    """
    """
    # Define the components
    u_cols = ['x_wind_speed_'+str(i) for i in range(12)]
    v_cols = ['y_wind_speed_'+str(i) for i in range(12)]
    w_cols = ['z_wind_speed_'+str(i) for i in range(12)]
    
    
    ws_cols = ['wind_speed_'+str(i) for i in range(12)]
    wd_cols = ['wind_direction_'+str(i) for i in range(12)]
    
    u  = lidar_data[u_cols]
    v  = lidar_data[v_cols]    
    w  = lidar_data[w_cols]    
    ws = lidar_data[ws_cols]        
    wd = lidar_data[wd_cols]    
    
    return u, v, w, ws, wd

def fit_curve(x,y,var):
    """
    Fit curve using a linear squares approach
    
    func can be: log (N*log(x)), 0 (D*x^0), 1 (C*x^1), 2 (B*x^2), 3 (A*x^3)
    """
    
    # Define the necessary matrices
    if var == 'wind':
        A_T = np.matrix([[np.log(i) for i in x],[i**2 for i in x],[i for i in x],[1 for i in x]])
    elif var == 'temp':
        A_T = np.matrix([[i**5 for i in x],[i**4 for i in x],[i**3 for i in x],[i**2 for i in x],[i for i in x],[1 for i in x]])  
    
    A = np.transpose(A_T)
    
    y = np.transpose(np.matrix([y]))
    
    # Estimate the regression-parameters using least-squares approach
    alpha = np.linalg.inv(A_T*A)*A_T*y
    
    return alpha
    
    
def estm_new_curve(new_x, alpha, var): 
    """
    Estimate new profile unsing estimated regression parameters
    """
    # Calculated fit and estimate corresponding derivative  
    if var == 'wind':
        A_T_new = np.matrix([[np.log(i) for i in new_x],[i**2 for i in new_x],[i for i in new_x],[1 for i in new_x]])#,[1 for i in new_z]])
#        dydx = np.array(alpha[0])[0,0]*1/new_x + np.array(alpha[1])[0,0]*3*new_x**2 + np.array(alpha[2])[0,0]*2*new_x +np.array(alpha[3])[0,0]
        dydx = np.array(alpha[0])[0,0]*1/new_x + np.array(alpha[1])[0,0]*2*new_x + np.array(alpha[2])[0,0] #+ 2* np.log(new_x)/new_x*np.array(alpha[4])[0,0]#+ 3*new_x**2*np.array(alpha[4])[0,0]
    elif var == 'temp':
        A_T_new = np.matrix([[i**5 for i in new_x],[i**4 for i in new_x],[i**3 for i in new_x],[i**2 for i in new_x],[i for i in new_x],[1 for i in new_x]]) 
        dydx = np.array(alpha[0])[0,0]*5*new_x**4 +np.array(alpha[1])[0,0]*4*new_x**3 +np.array(alpha[2])[0,0]*3*new_x**2 + np.array(alpha[3])[0,0]*2*new_x + np.array(alpha[4])[0,0]
    A_new = np.transpose(A_T_new)
    y_new = A_new*alpha
      
          
    return np.array(y_new)[:,0], dydx

# =============================================================================
# Processing Tools
# =============================================================================





from distutils.core import setup

setup(
  name = 'oblopy',        
  packages = ['oblopy'],   # Chose the same as "name"
  version = '0.0.1',      # Start with a small number and increase it with every change you make
  license='bsd-3-clause',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
  description = 'Python toolbox for reading, analysing and processing profiling Lidar observations.',   # Give a short description about your library
  author = 'Christiane Duscha',                   # Type in your name
  author_email = 'christiane.duscha@uib.no',      # Type in your E-Mail
  url = 'https://github.com/cdu022/oblopy',   # Provide either the link to your github or to your website
  download_url = 'https://github.com/cdu022/oblopy/archive/refs/tags/v0.0.1.tar.gz',   
  keywords = ['Offshore Boundary Layer Observatory (OBLO)','Marine Boundary Layer', 'Lidar','Retrieval','Motion correction','meteorology', 'bergen'],   # Keywords that define your package best
  install_requires=['matplotlib','numpy','scipy','pandas','netCDF4','cartopy','sklearn','glob', 'os','copy','datetime'],
  classifiers=[
    'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: BSD License',   # Again, pick a license
    'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
  ],
)